/***************************************************************************
 * Copyright 2012 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.service.impl;

import java.util.ArrayList;
import java.util.List;

import org.springframework.stereotype.Service;

import kieker.webgui.common.exception.InvalidInputSizeException;
import kieker.webgui.common.exception.UninitializedGraphException;
import kieker.webgui.service.IGraphLayoutService;

import org.eclipse.emf.common.util.EList;

import de.cau.cs.kieler.core.alg.BasicProgressMonitor;
import de.cau.cs.kieler.core.alg.IKielerProgressMonitor;
import de.cau.cs.kieler.core.kgraph.KEdge;
import de.cau.cs.kieler.core.kgraph.KLabeledGraphElement;
import de.cau.cs.kieler.core.kgraph.KNode;
import de.cau.cs.kieler.core.kgraph.KPort;
import de.cau.cs.kieler.kiml.AbstractLayoutProvider;
import de.cau.cs.kieler.kiml.klayoutdata.KEdgeLayout;
import de.cau.cs.kieler.kiml.klayoutdata.KPoint;
import de.cau.cs.kieler.kiml.klayoutdata.KShapeLayout;
import de.cau.cs.kieler.kiml.options.EdgeRouting;
import de.cau.cs.kieler.kiml.options.LayoutOptions;
import de.cau.cs.kieler.kiml.options.PortConstraints;
import de.cau.cs.kieler.kiml.util.KimlUtil;
import de.cau.cs.kieler.klay.layered.LayeredLayoutProvider;

/**
 * This class provides a single method, that is used for GraphFlow layouting. The JavaScript GraphFlow throws an "autoLayout"-Event upon using the function
 * autoLayout(). This event provides two Strings which may be used by this class' layoutGraph(nodes, edges) method. The return value can then be used as an argument
 * for the GraphFlow's loadPositionsFromLayout(positions) function.
 * 
 * @author row, Nils Christian Ehmke
 */
@Service
public final class GraphLayoutServiceImpl implements IGraphLayoutService {

	/**
	 * Default constructor. <b>Do not use this constructor. This bean is Spring managed.</b>
	 */
	public GraphLayoutServiceImpl() {
		// No code necessary
	}

	/**
	 * Parses a String array, building a graph with nodes and fixed ports.
	 * 
	 * @param positions
	 *            The node positions.
	 * @param layoutInformation
	 *            The necessary information about the graph.
	 * @throws InvalidInputSizeException
	 *             If the given input is somehow invalid in size.
	 */
	private static void addNodes(final String[] positions, final LayoutInformation layoutInformation) throws InvalidInputSizeException {
		layoutInformation.getGraph().getData(KShapeLayout.class).setProperty(LayoutOptions.EDGE_ROUTING, EdgeRouting.ORTHOGONAL);

		// prepare array of lists
		List<KLabeledGraphElement> child;

		final int dimHalf = Integer.parseInt(positions[0]);
		final int dim = dimHalf * 2;

		// check input Array for correct length
		if (((positions.length - 1) % 8) != 0) {
			throw new InvalidInputSizeException();
		}

		// loop variables
		KNode node;
		KPort port;
		KShapeLayout layout;
		int width;
		int height;
		int portCount;
		int portX;
		int portY;
		int c = 1;
		int i = 0;

		// iterate input node positions
		while (c < positions.length) {
			// get size and increment array pointer
			width = Integer.parseInt(positions[c++]);
			height = Integer.parseInt(positions[c++]);

			// create Node
			node = KimlUtil.createInitializedNode();
			node.setParent(layoutInformation.getGraph());

			child = new ArrayList<KLabeledGraphElement>();
			layoutInformation.getChildren().add(i, child);
			child.add(node);

			// set node dimensions
			layout = node.getData(KShapeLayout.class);
			layout.setWidth(width);
			layout.setHeight(height);
			layout.setProperty(LayoutOptions.FIXED_SIZE, Boolean.TRUE);
			layout.setProperty(LayoutOptions.PORT_CONSTRAINTS, PortConstraints.FIXED_POS);
			portX = -dimHalf;

			// add Ports
			for (int portType = 0; portType < 3; portType++) {
				// determine port x-position
				if (portType == 1) {
					portX = width - dimHalf;
				}

				portCount = Integer.parseInt(positions[c++]);
				portY = Integer.parseInt(positions[c++]);

				final int lp = portCount;
				for (int n = 0; n < lp; n++) {

					// create port
					port = KimlUtil.createInitializedPort();
					port.setNode(node);
					child.add(port);

					// set layout for Port
					layout = port.getData(KShapeLayout.class);

					// set position and dimension
					layout.setWidth(dim);
					layout.setHeight(dimHalf);
					layout.setYpos(portY);
					layout.setXpos(portX);

					portY += dim;
				}
			}
			i++;
		}
	}

	/**
	 * Parses an array which must contain quadruples of numbers.<br/>
	 * The first number must be the index of the children-array, where
	 * the source node is located.<br/>
	 * The second number must be the index of the children-array, where
	 * the target node is located.<br/>
	 * The third number must be the index of the source node ArrayList, where
	 * the source port is located.<br/>
	 * The fourth number must be the index of the target node ArrayList, where
	 * the target port is located.
	 * 
	 * @param edgeInfo
	 *            The single elements for the edges.
	 * @param layoutInformation
	 *            The necessary information about the graph.
	 * @throws UninitializedGraphException
	 *             If the graph has not yet been initialized.
	 * @throws InvalidInputSizeException
	 *             If the input size is invalid.
	 */
	private static void addEdges(final String[] edgeInfo, final LayoutInformation layoutInformation) throws UninitializedGraphException, InvalidInputSizeException {
		// cannot add edges if no graph was constructed
		if (layoutInformation.getChildren() == null) {
			throw new UninitializedGraphException();
		}
		if ((edgeInfo.length % 4) != 0) {
			throw new InvalidInputSizeException();
		}

		// set up loop variables
		KEdge edge;
		int sourceID;
		int targetID;
		KPort sourcePort;
		KPort targetPort;
		int e = 0;
		final int l = edgeInfo.length;
		int ea = 0;

		while (e < l) {
			// read from String
			sourceID = Integer.parseInt(edgeInfo[e++]);
			targetID = Integer.parseInt(edgeInfo[e++]);
			sourcePort = (KPort) layoutInformation.getChildren().get(sourceID).get(1 + Integer.parseInt(edgeInfo[e++]));
			targetPort = (KPort) layoutInformation.getChildren().get(targetID).get(1 + Integer.parseInt(edgeInfo[e++]));

			// add edge to graph
			edge = KimlUtil.createInitializedEdge();
			edge.setSource((KNode) layoutInformation.getChildren().get(sourceID).get(0));
			edge.setTarget((KNode) layoutInformation.getChildren().get(targetID).get(0));

			// set ports of edge
			edge.setSourcePort(sourcePort);
			sourcePort.getEdges().add(edge);
			edge.setTargetPort(targetPort);
			targetPort.getEdges().add(edge);

			layoutInformation.getEdges().add(ea++, edge);
		}
	}

	/**
	 * This simple method "corrects" an array containing only an empty string.
	 * 
	 * @param arr
	 *            The array to be corrected.
	 * @return An corrected version of the array.
	 */
	private static String[] correctEmptyArray(final String[] arr) {
		return ((arr.length == 1) && arr[0].isEmpty()) ? new String[0] : arr;
	}

	/**
	 * Assembles the necessary layout information using the given parameters.
	 * 
	 * @param nodesStr
	 *            The string containing the information about the nodes.
	 * @param edgesStr
	 *            The string containing the information about the edges.
	 * @return An object containing the layout information for the graph.
	 * @throws UninitializedGraphException
	 *             If the graph has not been initialized yet.
	 * @throws InvalidInputSizeException
	 *             IF the input size of the graph is somehow invalid.
	 */
	private static LayoutInformation assembleLayoutInformation(final String nodesStr, final String edgesStr) throws UninitializedGraphException,
			InvalidInputSizeException {
		final String[] edgeInfo = GraphLayoutServiceImpl.correctEmptyArray(edgesStr.split(" "));
		final String[] positions = GraphLayoutServiceImpl.correctEmptyArray(nodesStr.split(" "));

		final List<List<KLabeledGraphElement>> children = new ArrayList<List<KLabeledGraphElement>>(positions.length / 2);
		final List<KEdge> edges = new ArrayList<KEdge>(edgeInfo.length / 4);
		final LayoutInformation layoutInformation = new LayoutInformation(KimlUtil.createInitializedNode(), children, edges);

		GraphLayoutServiceImpl.addNodes(positions, layoutInformation);
		GraphLayoutServiceImpl.addEdges(edgeInfo, layoutInformation);

		return layoutInformation;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.service.impl.IGraphLayoutService#layoutGraph(java.lang.String, java.lang.String)
	 */
	@Override
	public String layoutGraph(final String nodes, final String edges) throws UninitializedGraphException, InvalidInputSizeException {
		final LayoutInformation layoutInformation = GraphLayoutServiceImpl.assembleLayoutInformation(nodes, edges);

		// Create a progress monitor
		final IKielerProgressMonitor progressMonitor = new BasicProgressMonitor();

		// Create the layout provider
		final AbstractLayoutProvider layoutProvider = new LayeredLayoutProvider();

		// Perform layout on the created graph
		layoutProvider.doLayout(layoutInformation.getGraph(), progressMonitor);

		// convert layouted graph positions to String
		final String layoutedGraph = GraphLayoutServiceImpl.getPositions(layoutInformation);
		final String bendPoints = GraphLayoutServiceImpl.getBendPoints(layoutInformation);
		return layoutedGraph + "#" + bendPoints;
	}

	/**
	 * Reads the (layouted) positions of a constructed KGraph and writes them as integers to a string, seperating each number with a space.
	 * 
	 * @param layoutInformation
	 *            The object containing the necessary information about the graph.
	 * @return The constructed string.
	 */
	private static String getPositions(final LayoutInformation layoutInformation) {
		final StringBuilder sb = new StringBuilder();
		KShapeLayout layout;

		for (final KNode node : layoutInformation.getGraph().getChildren()) {
			layout = node.getData(KShapeLayout.class);
			final int x = Math.round(layout.getXpos() + (layout.getWidth() / 2));
			final int y = Math.round(layout.getYpos() + (layout.getHeight() / 2));
			sb.append(x);
			sb.append(" ");
			sb.append(y);
			sb.append(" ");
		}

		return sb.toString();
	}

	/**
	 * This method iterates through all edges of the given object and creates a semicolon separated String containing bend points of all edges.
	 * 
	 * @param layoutInformation
	 *            The object containing the information about the graph and the layout.
	 * @return The string containing the bend points.
	 */
	private static String getBendPoints(final LayoutInformation layoutInformation) {
		final StringBuilder sb = new StringBuilder();

		// iterate edges
		for (final KEdge edge : layoutInformation.getEdges()) {
			final KEdgeLayout layout = edge.getData(KEdgeLayout.class);
			final EList<KPoint> bendPoints = layout.getBendPoints();

			// iterate bend points
			final int bl = bendPoints.size();
			for (int b = 0; b < bl; b++) {
				final KPoint point = bendPoints.get(b);
				sb.append(Math.round(point.getX()));
				sb.append(" ");
				sb.append(Math.round(point.getY()));
				sb.append(" ");
			}
			sb.append("; ");
		}

		return sb.toString();
	}

	/**
	 * A container for the layout information which will be assembled during the layout process.
	 * 
	 * @author Nils Christian Ehmke
	 */
	private static class LayoutInformation {

		private final KNode graph;
		/**
		 * An array containing the child node and the ports from left to right.
		 */
		private final List<List<KLabeledGraphElement>> children;
		/**
		 * An array containing edges. This way we can return edge information in the same order.
		 */
		private final List<KEdge> edges;

		/**
		 * Creates a new instance of this class.
		 * 
		 * @param graph
		 *            The graph itself.
		 * @param children
		 *            The child notes and the ports.
		 * @param edges
		 *            The edges of the graph.
		 */
		public LayoutInformation(final KNode graph, final List<List<KLabeledGraphElement>> children, final List<KEdge> edges) {
			this.graph = graph;
			this.children = children;
			this.edges = edges;
		}

		/**
		 * Getter for the property {@link LayoutInformation#graph}.
		 * 
		 * @return The current value of the property.
		 */
		public KNode getGraph() {
			return this.graph;
		}

		/**
		 * Getter for the property {@link LayoutInformation#children}.
		 * 
		 * @return The current value of the property.
		 */
		public List<List<KLabeledGraphElement>> getChildren() {
			return this.children;
		}

		/**
		 * Getter for the property {@link LayoutInformation#edges}.
		 * 
		 * @return The current value of the property.
		 */
		public List<KEdge> getEdges() {
			return this.edges;
		}

	}
}
